#define MAX_INSTRUMENT_GUI	64

enum {
	INSGUI_SIGNED_BYTE = 0,
	INSGUI_UNSIGNED_BYTE,
	INSGUI_SIGNED_WORD,
	INSGUI_UNSIGNED_WORD,
	INSGUI_HEX_STRING,
	INSGUI_OPTION,
	INSGUI_LABEL,
	INSGUI_ENVELOPE,
	INSGUI_SELECT_FILE,
	INSGUI_TEXT_STRING
};

struct instrumentGuiStruct {
	int type;
	int min;	//min value for a numeric input
	int max;	//max value for a numeric input
	int digits;
	int byte;
	int max_len;	//max length of a string or envelope input
	int option_count;
	int y;
	int res;

	char caption[1024];
	char option[128];

	char file_ext[8];
	int file_id;
	int file_save;	//save mode flag
};

struct instrumentGuiStruct instrumentGuiList[MAX_INSTRUMENT_GUI];

int instrumentGuiCount;
int instrumentGuiCursor=0;
int instrumentGuiNumber;
int instrumentGuiParamXOff;
int instrumentGuiActiveControls;

bool instrumentGuiEnvelope;
int instrumentGuiEnvelopeCursor;
int instrumentGuiEnvelopeOff;
bool instrumentGuiDump = false;



void instrument_gui_add_number(int type,const char* caption,int min,int max,int byte)
{
	unsigned int i = instrumentGuiCount;

	instrumentGuiList[i].type=type;
	instrumentGuiList[i].min=min;
	instrumentGuiList[i].max=max;

	if(min>max) max=min;

	unsigned int digits = 1;

	if(max>=10)    digits=2;
	if(max>=100)   digits=3;
	if(max>=1000)  digits=4;
	if(max>=10000) digits=5;

	instrumentGuiList[i].digits=digits;
	instrumentGuiList[i].byte=byte;

	strncpy(instrumentGuiList[i].caption, caption, sizeof(instrumentGuiList[i].caption));

	++instrumentGuiCount;
	++instrumentGuiActiveControls;
}



void instrument_gui_add_envelope(const char* caption, int max_len, int max_val, int res, int byte, bool env)
{
	unsigned int i = instrumentGuiCount;

	instrumentGuiList[i].type = !env ? INSGUI_HEX_STRING : INSGUI_ENVELOPE;
	instrumentGuiList[i].max_len = max_len;
	instrumentGuiList[i].min = 0;
	instrumentGuiList[i].max = max_val;
	instrumentGuiList[i].byte = byte;
	instrumentGuiList[i].res = res;

	strncpy(instrumentGuiList[i].caption, caption, sizeof(instrumentGuiList[i].caption));

	++instrumentGuiCount;
	++instrumentGuiActiveControls;
}



void instrument_gui_add_hex_string(const char* caption,int max_len,int res,int byte,bool env)
{
	unsigned int i = instrumentGuiCount;

	instrumentGuiList[i].type = !env ? INSGUI_HEX_STRING : INSGUI_ENVELOPE;
	instrumentGuiList[i].max_len = max_len;
	instrumentGuiList[i].byte = byte;
	instrumentGuiList[i].res = res;

	strncpy(instrumentGuiList[i].caption, caption, sizeof(instrumentGuiList[i].caption));

	++instrumentGuiCount;
	++instrumentGuiActiveControls;
}



void instrument_gui_add_text_string(const char* caption, int max_len, int res, int byte)
{
	unsigned int i = instrumentGuiCount;

	instrumentGuiList[i].type = INSGUI_TEXT_STRING;
	instrumentGuiList[i].max_len = max_len;
	instrumentGuiList[i].byte = byte;
	instrumentGuiList[i].res = res;

	strncpy(instrumentGuiList[i].caption, caption, sizeof(instrumentGuiList[i].caption));

	++instrumentGuiCount;
	++instrumentGuiActiveControls;
}



void instrument_gui_add_option(const char* caption,const char* option,int byte)
{
	unsigned int n = 1;

	for (unsigned int i = 0; i < strlen(option); ++i) if (option[i] == '|') ++n;

	unsigned int i = instrumentGuiCount;

	instrumentGuiList[i].type=INSGUI_OPTION;
	instrumentGuiList[i].byte=byte;
	instrumentGuiList[i].option_count=n;

	strncpy(instrumentGuiList[i].option, option, sizeof(instrumentGuiList[i].option));
	strncpy(instrumentGuiList[i].caption, caption, sizeof(instrumentGuiList[i].caption));

	++instrumentGuiCount;
	++instrumentGuiActiveControls;
}



void instrument_gui_add_label(const char* caption)
{
	unsigned int i = instrumentGuiCount;

	instrumentGuiList[i].type=INSGUI_LABEL;

	strncpy(instrumentGuiList[i].caption, caption, sizeof(instrumentGuiList[i].caption));

	++instrumentGuiCount;
}



void instrument_gui_add_select_file(const char* caption, const char* file_ext, int file_save, int file_id)
{
	unsigned int i = instrumentGuiCount;

	instrumentGuiList[i].type = INSGUI_SELECT_FILE;

	strncpy(instrumentGuiList[i].caption, caption, sizeof(instrumentGuiList[i].caption));
	strncpy(instrumentGuiList[i].file_ext, file_ext, sizeof(instrumentGuiList[i].file_ext));
	instrumentGuiList[i].file_id = file_id;
	instrumentGuiList[i].file_save = file_save;
	++instrumentGuiCount;
	++instrumentGuiActiveControls;
}



void set_instrument_editor(void)
{
	memset(instrumentGuiList, 0, sizeof(instrumentGuiList));

	instrumentGuiCount = 0;
	++instrumentGuiActiveControls = 0;
	instrumentGuiEnvelope = false;
	instrumentGuiEnvelopeCursor = 0;

	change_instrument();

	int len = 0;
	int y = 0;

	for (int i = 0; i < instrumentGuiCount; ++i)
	{
		instrumentGuiList[i].y = y;

		if (instrumentGuiList[i].type != INSGUI_ENVELOPE)
		{
			y += get_str_height(instrumentGuiList[i].caption);
		}
		else
		{
			y += instrumentGuiList[i].max / instrumentGuiList[i].res;
		}

		if (instrumentGuiList[i].type == INSGUI_LABEL) continue;

		if ((int)strlen(instrumentGuiList[i].caption) > len)
		{
			len = strlen(instrumentGuiList[i].caption);
		}
	}

	if (instrumentGuiCount > 0)
	{
		while (instrumentGuiList[instrumentGuiCursor].type == INSGUI_LABEL && instrumentGuiCursor < instrumentGuiCount) ++instrumentGuiCursor;

		if (instrumentGuiCursor >= instrumentGuiCount)
		{
			instrumentGuiCursor = instrumentGuiCount - 1;
			while (instrumentGuiList[instrumentGuiCursor].type == INSGUI_LABEL && instrumentGuiCursor > 0) --instrumentGuiCursor;
		}
	}

	instrumentGuiParamXOff=PATTERN_XPOS+1+len+2;

	set_edit_mode(MODE_INSTRUMENT_EDITOR);
}



int instrument_get_value(int i)
{
	unsigned char* data=module.instrument[curInstrument].data+instrumentGuiList[i].byte;

	int n = 0;

	switch(instrumentGuiList[i].type)
	{
	case INSGUI_SIGNED_BYTE:   n = *((char*)data); break;
	case INSGUI_UNSIGNED_BYTE: n = *((unsigned char*)data); break;
	case INSGUI_SIGNED_WORD:   n = *((short*)data); break;
	case INSGUI_UNSIGNED_WORD: n = *((unsigned short*)data); break;
	}

	int min = instrumentGuiList[i].min;
	int max = instrumentGuiList[i].max;

	if (n < min) n = min;
	if (n > max) n = max;

	return n;
}



void instrument_set_value(int i,int n)
{
	unsigned char* data=module.instrument[curInstrument].data+instrumentGuiList[i].byte;

	int min=instrumentGuiList[i].min;
	int max=instrumentGuiList[i].max;

	if(n<min) n=min;
	if(n>max) n=max;

	switch(instrumentGuiList[i].type)
	{
	case INSGUI_SIGNED_BYTE:   *(char*)data=(char)n; break;
	case INSGUI_UNSIGNED_BYTE: *(unsigned char*)data=(unsigned char)n; break;
	case INSGUI_SIGNED_WORD:   *(short*)data=(short)n; break;
	case INSGUI_UNSIGNED_WORD: *(unsigned short*)data=(unsigned short)n; break;
	}
}



void instrument_add_value(int value)
{
	int i = instrumentGuiCursor;

	unsigned char* data = module.instrument[curInstrument].data + instrumentGuiList[i].byte;

	switch (instrumentGuiList[i].type)
	{
	case INSGUI_SIGNED_BYTE:
	case INSGUI_UNSIGNED_BYTE:
	case INSGUI_SIGNED_WORD:
	case INSGUI_UNSIGNED_WORD:
		{
			instrument_set_value(i, instrument_get_value(i) + value);
		}
		break;

	case INSGUI_OPTION:
		{
			if(value<0)
			{
				if (*data > 0)--*data; else *data = instrumentGuiList[i].option_count - 1;
			}
			else
			{
				if (*data < instrumentGuiList[i].option_count - 1)++*data; else *data = 0;
			}
		}
		break; 
	}

	set_instrument_editor();
}



void draw_instrument_editor(void)
{
	char str[1024];

	for (int i = 0; i < textScreenWidth; ++i)
	{
		PUT_ATTR(i,0,COL_HEADER);
		PUT_ATTR(i,textScreenHeight-1,COL_HEADER);
	}

	put_str(1,0,"Instrument editor",COL_HEADER);

	if(!module.instruments)
	{
		put_str(PATTERN_XPOS+1,PATTERN_YPOS+1,"This engine does not support instruments",COL_PROGRAM_BACKGROUND);

		return;
	}

	put_str(1,textScreenHeight-1,"Instrument     Channel --  ~F2~Save~F3~Load~F4~Back~F5~PlayRow",COL_HEADER);
	put_num(12,textScreenHeight-1,curInstrument+1,3);

	if(module.columnChannel[curColumn]>=0) put_num(24,textScreenHeight-1,module.columnChannel[curColumn]+1,2);

	int x1 = PATTERN_XPOS + 1;
	int x2 = instrumentGuiParamXOff + 10;
	int lowest_y = 0;

	for (int ins = 0; ins < instrumentGuiCount; ++ins)
	{
		int y = PATTERN_YPOS + 1 + instrumentGuiList[ins].y;

		if (y > lowest_y) lowest_y = y;

		unsigned char col = (ins == instrumentGuiCursor) ? COL_CURSOR_ROW : COL_PROGRAM_BACKGROUND;

		if (instrumentGuiList[ins].type == INSGUI_LABEL) col = COL_DIRECTORY;

		str[0] = 0;

		if (instrumentGuiList[ins].type != INSGUI_LABEL)
		{
			strncpy(str, " ", sizeof(str));
			strncat(str, instrumentGuiList[ins].caption, sizeof(str));
			strncat(str, " ", sizeof(str));
			put_str(x1 - 1, y, str, col);
		}
		else
		{
			put_str_wrap(x1, y, textScreenWidth - x1, -1, instrumentGuiList[ins].caption, col, 0);
		}
		
		int d = instrumentGuiList[ins].digits;

		unsigned char* data = &module.instrument[curInstrument].data[instrumentGuiList[ins].byte];
		
		int min = instrumentGuiList[ins].min;
		int max = instrumentGuiList[ins].max;

		switch(instrumentGuiList[ins].type)
		{
		case INSGUI_SIGNED_BYTE:
		case INSGUI_SIGNED_WORD:
			{
				if (instrument_get_value(ins)<0) PUT_CHAR(instrumentGuiParamXOff - 1, y, '-');
			}
			//break;	//intentionally, to fall through to the next cases
		case INSGUI_UNSIGNED_BYTE:
		case INSGUI_UNSIGNED_WORD:
			{
				put_num(instrumentGuiParamXOff, y, abs(instrument_get_value(ins)), d);

				snprintf(str, sizeof(str), "(%i..%i)", min, max);

				put_str(x2, y, str, COL_PROGRAM_BACKGROUND);
			}
			break;

		case INSGUI_HEX_STRING:
			{
			for (int j = 0; j < instrumentGuiList[ins].max_len; ++j)
				{
					PUT_CHAR(instrumentGuiParamXOff+j,y,NUM2HEX(data[j]));

					if ((instrumentGuiParamXOff + j) >= textScreenWidth - 3) break;
				}
			}
			break;

		case INSGUI_TEXT_STRING:
		{
			for (int j = 0; j < instrumentGuiList[ins].max_len; ++j)
			{
				char c = data[j];
				if (c < 0x20) break;
				PUT_CHAR(instrumentGuiParamXOff + j, y, c);

				if ((instrumentGuiParamXOff + j) >= textScreenWidth - 3) break;
			}
		}
		break;

		case INSGUI_OPTION:
			{
				char* option = instrumentGuiList[ins].option;

				int k = *data;

				if(k)
				{
					for (int j = 0; j < (int)strlen(option); ++j)
					{
						if (option[j] == '|')
						{
							--k;

							if(!k)
							{
								option = &option[j + 1];
								break;
							}
						}
					}
				}

				for (int j = 0; j < (int)strlen(option); ++j)
				{
					char c = option[j];

					if(c=='|') break;

					PUT_CHAR(instrumentGuiParamXOff+j,y,c);
				}
			}
			break;

		case INSGUI_ENVELOPE:
			{
				int len = instrumentGuiList[ins].max_len;

				if (instrumentGuiParamXOff + len >= (textScreenWidth - 2))
				{
					len = textScreenWidth - 2 - instrumentGuiParamXOff;
					instrumentGuiEnvelopeOff = instrumentGuiEnvelopeCursor - len / 2;
					if (instrumentGuiEnvelopeOff < 0) instrumentGuiEnvelopeOff = 0;
					if (instrumentGuiEnvelopeOff > (instrumentGuiList[ins].max_len - len)) instrumentGuiEnvelopeOff = (instrumentGuiList[ins].max_len - len);

					if (instrumentGuiEnvelopeOff < (instrumentGuiList[ins].max_len - len))
					{
						for (int i = 0; i < instrumentGuiList[ins].max; ++i)
						{
							PUT_CHAR_ATTR((textScreenWidth - 2), y + i, '>', COL_PROGRAM_BACKGROUND);
						}
					}
				}
				else
				{
					instrumentGuiEnvelopeOff = 0;
				}

				if (instrumentGuiEnvelopeOff > 0)
				{
					for (int i = 0; i < instrumentGuiList[ins].max; ++i)
					{
						PUT_CHAR_ATTR(instrumentGuiParamXOff - 1, y + i, '<', COL_PROGRAM_BACKGROUND);
					}
				}

				for (int j = 0; j < len; ++j)
				{
					int off = j + instrumentGuiEnvelopeOff;
					int val = module.instrument[curInstrument].data[instrumentGuiList[ins].byte + off];

					if (instrumentGuiList[ins].res == 1)
					{
						for (int k = 0; k < instrumentGuiList[ins].max; ++k)
						{
							char c = 4;

							if (k <= val) c = 5;

							PUT_CHAR_ATTR(instrumentGuiParamXOff + j, y + (instrumentGuiList[ins].max - 1 - k), c, (instrumentGuiEnvelope && (off == instrumentGuiEnvelopeCursor)) ? COL_CURSOR : COL_PROGRAM_BACKGROUND);
						}
					}

					if (instrumentGuiList[ins].res == 2)
					{
						for (int k = 0; k < instrumentGuiList[ins].max; k += 2)
						{
							char c = 0;

							if (k <= val) c = 1;
							if (k < val)  c = 2;

							PUT_CHAR_ATTR(instrumentGuiParamXOff + j, y + (instrumentGuiList[ins].max - 1 - k) / 2, c, (instrumentGuiEnvelope && (off == instrumentGuiEnvelopeCursor)) ? COL_CURSOR : COL_PROGRAM_BACKGROUND);
						}
					}

					if (instrumentGuiEnvelope && (((off % curPageStep) == 0) || (off == instrumentGuiList[ins].max_len - 1)))
					{
						PUT_CHAR_ATTR(instrumentGuiParamXOff + j, y + instrumentGuiList[ins].max, '\'', COL_PROGRAM_BACKGROUND);
					}

					if (instrumentGuiEnvelope && (off == instrumentGuiEnvelopeCursor))
					{
						PUT_CHAR_ATTR(instrumentGuiParamXOff + j, y + instrumentGuiList[ins].max, '^', COL_PROGRAM_BACKGROUND);
					}
				}

				int hgt = instrumentGuiList[ins].max;

				if (instrumentGuiList[ins].res) hgt /= 2;

				if ((y + hgt) > lowest_y) lowest_y = y + hgt;
			}
			break;
		}
	}

	if (instrumentGuiDump)
	{
		int x = 2; 
		int y = lowest_y + 4;
		
		for (int i = 0; i < MAX_INSTRUMENT_DATA; ++i)
		{
			unsigned char n = module.instrument[curInstrument].data[i];

			char c1 = '.';
			char c2 = '.';

			if (n)
			{
				c1 = NUM2HEX(n >> 4); if (c1 == '.') c1 = '0';
				c2 = NUM2HEX(n & 0x0f); if (c2 == '.') c2 = '0';
			}

			PUT_CHAR_ATTR(x + 0, y, c1, COL_MARKER_LIST);
			PUT_CHAR_ATTR(x + 1, y, c2, COL_MARKER_LIST);
			
			x += 2;

			if (x >= textScreenWidth - 2)
			{
				x = 2;
				++y;
				if (y >= textScreenHeight - 2) break;
			}
		}
	}
}



void move_envelope_cursor(int dx)
{
	if (dx < 0 && instrumentGuiEnvelopeCursor == 0)
	{
		instrumentGuiEnvelopeCursor = instrumentGuiList[instrumentGuiCursor].max_len - 1;
		return;
	}

	if (dx > 0 && instrumentGuiEnvelopeCursor == (instrumentGuiList[instrumentGuiCursor].max_len - 1))
	{
		instrumentGuiEnvelopeCursor = 0;
		return;
	}

	instrumentGuiEnvelopeCursor += dx;

	if (instrumentGuiEnvelopeCursor < 0) instrumentGuiEnvelopeCursor = 0;

	if (instrumentGuiEnvelopeCursor >= instrumentGuiList[instrumentGuiCursor].max_len) instrumentGuiEnvelopeCursor = instrumentGuiList[instrumentGuiCursor].max_len - 1;
}



void instrument_editor_events(SDL_Event event)
{
	if (editMode == MODE_INSTRUMENT_INPUT_DONE)
	{
		instrument_set_value(instrumentGuiCursor, instrumentGuiNumber);
	}

	if (editMode != MODE_INSTRUMENT_EDITOR)
	{
		set_instrument_editor(); 
		
		set_edit_mode(MODE_INSTRUMENT_EDITOR);
	}

	switch(event.type)
	{
	case SDL_KEYDOWN:
		{
			if(!instrumentGuiEnvelope)
			{
				//instrument editor input

				//first buttons that is always active

				switch (event.key.keysym.sym)
				{
				case SDLK_ESCAPE:
				case SDLK_F4:
				{
					set_edit_mode(MODE_EDIT_SONG);
				}
				return;

				case SDLK_F5:
				{
					music_play_row();
				}
				break;

				case SDLK_EQUALS:
				case SDLK_KP_PLUS:
				{
					if (curInstrument<module.instruments - 1) ++curInstrument;

					set_instrument_editor();
				}
				break;

				case SDLK_MINUS:
				case SDLK_KP_MINUS:
				{
					if (curInstrument) --curInstrument;

					set_instrument_editor();
				}
				break;

				case SDLK_TAB://move cursor to the next or previous channel
				{
					if (event.key.keysym.mod&(KMOD_LSHIFT | KMOD_RSHIFT)) change_channel(-1); else change_channel(1);

					set_instrument_editor();
				}
				break;
				}

				//then buttons that only active when there is instruments support and active controls in the list

				if (!module.instruments || !instrumentGuiActiveControls) return;

				switch(event.key.keysym.sym)
				{
				case SDLK_UP:
					{
						while(1)
						{
							--instrumentGuiCursor;

							if(instrumentGuiCursor<0) instrumentGuiCursor=instrumentGuiCount-1;

							if(instrumentGuiList[instrumentGuiCursor].type!=INSGUI_LABEL) break;
						}
					}
					break;

				case SDLK_DOWN:
					{
						while(1)
						{
							++instrumentGuiCursor;

							if(instrumentGuiCursor>=instrumentGuiCount) instrumentGuiCursor=0;

							if(instrumentGuiList[instrumentGuiCursor].type!=INSGUI_LABEL) break;
						}
					}
					break;

				case SDLK_PAGEUP:
					{
						instrumentGuiCursor = 0;

						while (instrumentGuiList[instrumentGuiCursor].type == INSGUI_LABEL) ++instrumentGuiCursor;
					}
					break;

				case SDLK_PAGEDOWN:
					{
						instrumentGuiCursor = instrumentGuiCount - 1;

						while (instrumentGuiList[instrumentGuiCursor].type == INSGUI_LABEL) --instrumentGuiCursor;
					}
					break;

				case SDLK_F2:
					{
					set_select_file("Save instrument file", instrumentDirectory, sizeof(instrumentDirectory), MODE_INSTRUMENT_EDITOR_RETURN, INSTRUMENT_EXTENSION, true, EVENT_SAVE_INSTRUMENT, selectFileInstrumentLast, sizeof(selectFileInstrumentLast), 1, 0);
					}
					break;

				case SDLK_F3:
					{
					set_select_file("Load instrument file", instrumentDirectory, sizeof(instrumentDirectory), MODE_INSTRUMENT_EDITOR_RETURN, INSTRUMENT_EXTENSION, true, EVENT_LOAD_INSTRUMENT, selectFileInstrumentLast, sizeof(selectFileInstrumentLast), 0, 0);
					}
					break;

				case SDLK_LEFT:
					{
						instrument_add_value(-1);
					}
					break;

				case SDLK_RIGHT:
					{
						instrument_add_value(1);
					}
					break;

				case SDLK_RETURN:
					{
						switch(instrumentGuiList[instrumentGuiCursor].type)
						{
							case INSGUI_HEX_STRING:
							{
								set_input_hex_string(instrumentGuiParamXOff,instrumentGuiList[instrumentGuiCursor].y+PATTERN_YPOS+1,instrumentGuiList[instrumentGuiCursor].max_len,MODE_INSTRUMENT_INPUT_DONE,(char*)&module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte]);
							}
							break;

							case INSGUI_TEXT_STRING:
							{
								set_input_string(instrumentGuiParamXOff, instrumentGuiList[instrumentGuiCursor].y + PATTERN_YPOS + 1, instrumentGuiList[instrumentGuiCursor].max_len, MODE_INSTRUMENT_INPUT_DONE, false, (char*)&module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte], NULL);
							}
							break;

							case INSGUI_ENVELOPE:
							{
								instrumentGuiEnvelope=true;
								instrumentGuiEnvelopeOff = 0;

								if (instrumentGuiEnvelopeCursor > instrumentGuiList[instrumentGuiCursor].max_len - 1) instrumentGuiEnvelopeCursor = instrumentGuiList[instrumentGuiCursor].max_len - 1;
							}
							break;

							case INSGUI_OPTION:
							{
								instrument_add_value(1);
							}
							break;

							case INSGUI_SELECT_FILE:
							{
								set_select_file(instrumentGuiList[instrumentGuiCursor].caption, itemSelectFilename, sizeof(itemSelectFilename), MODE_INSTRUMENT_EDITOR_RETURN, instrumentGuiList[instrumentGuiCursor].file_ext, true, EVENT_SCRIPT_FILE_SELECT, selectFileInstrumentLast, sizeof(selectFileInstrumentLast), instrumentGuiList[instrumentGuiCursor].file_save, instrumentGuiList[instrumentGuiCursor].file_id);
							}
							break;

							default:
							{
								instrumentGuiNumber=instrument_get_value(instrumentGuiCursor);

								bool sign = false;

								switch(instrumentGuiList[instrumentGuiCursor].type)
								{
								case INSGUI_SIGNED_BYTE:
								case INSGUI_SIGNED_WORD:
									sign=true;
									break;
								}

								set_input_number(instrumentGuiParamXOff, instrumentGuiList[instrumentGuiCursor].y + PATTERN_YPOS + 1, instrumentGuiList[instrumentGuiCursor].digits, MODE_INSTRUMENT_INPUT_DONE, true, &instrumentGuiNumber, sign, NULL);
							}
						}					
					}
					break;

				case SDLK_BACKQUOTE:
					{
						instrumentGuiDump ^= true;
					}
					break;
				}
			}
			else
			{
				//envelope editor input

				switch(event.key.keysym.sym)
				{
					case SDLK_ESCAPE:
					case SDLK_RETURN:
					{
						instrumentGuiEnvelope = false;
						instrumentGuiEnvelopeOff = 0;
					}
					break;

					case SDLK_F5:
					{
						music_play_row();
					}
					break;

					case SDLK_UP:
					{
						int val = module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte + instrumentGuiEnvelopeCursor];

						++val;

						if (val >= instrumentGuiList[instrumentGuiCursor].max) val = 0;

						module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte+instrumentGuiEnvelopeCursor]=val;
					}
					break;

					case SDLK_DOWN:
					{
						int val = module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte + instrumentGuiEnvelopeCursor];

						--val;

						if (val < 0) val = instrumentGuiList[instrumentGuiCursor].max - 1;

						module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte+instrumentGuiEnvelopeCursor]=val;
					}
					break;

					case SDLK_LEFT:
					{
						move_envelope_cursor(-1);
					}
					break;

					case SDLK_RIGHT:
					{
						move_envelope_cursor(1);
					}
					break;

					case SDLK_INSERT:
					{
						int off = instrumentGuiList[instrumentGuiCursor].byte;

						for (int i = instrumentGuiList[instrumentGuiCursor].max_len - 1; i > instrumentGuiEnvelopeCursor; --i)
						{
							module.instrument[curInstrument].data[off+i]=module.instrument[curInstrument].data[off+i-1];
						}

						++instrumentGuiEnvelopeCursor;

						if (instrumentGuiEnvelopeCursor >= instrumentGuiList[instrumentGuiCursor].max_len - 1) instrumentGuiEnvelopeCursor = instrumentGuiList[instrumentGuiCursor].max_len - 1;
					}
					break;

					case SDLK_BACKSPACE:
					{
						if(instrumentGuiEnvelopeCursor)
						{
							int off = instrumentGuiList[instrumentGuiCursor].byte;

							for (int i = instrumentGuiEnvelopeCursor; i < instrumentGuiList[instrumentGuiCursor].max_len - 1; ++i)
							{
								module.instrument[curInstrument].data[off+i]=module.instrument[curInstrument].data[off+i+1];
							}

							module.instrument[curInstrument].data[off + instrumentGuiList[instrumentGuiCursor].max_len - 1] = 0;

							--instrumentGuiEnvelopeCursor;
						}
					}
					break;

					case SDLK_DELETE:
					{
						int off = instrumentGuiList[instrumentGuiCursor].byte;

						for (int i = instrumentGuiEnvelopeCursor; i < instrumentGuiList[instrumentGuiCursor].max_len - 1; ++i)
						{
							module.instrument[curInstrument].data[off+i]=module.instrument[curInstrument].data[off+i+1];
						}

						module.instrument[curInstrument].data[off + instrumentGuiList[instrumentGuiCursor].max_len - 1] = 0;
					}
					break;

					case SDLK_HOME:
					{
						instrumentGuiEnvelopeCursor = 0;
					}
					break;

					case SDLK_END:
					{
						instrumentGuiEnvelopeCursor = instrumentGuiList[instrumentGuiCursor].max_len - 1;
					}
					break;

					case SDLK_PAGEDOWN:
					{
						move_envelope_cursor(curPageStep);
					}
					break;

					case SDLK_PAGEUP:
					{
						move_envelope_cursor(-curPageStep);
					}
					break;
				}
			}
		}
		break;
	}
}