vector<string> selectEngineFileList = vector<string>();
vector<string> selectEngineTitleList = vector<string>();
vector<string> selectEngineAboutList = vector<string>();

int selectEngineCount;
int selectEngineCursor;

char selectEngineTitle[128];
char selectEngineAbout[MAX_TEXT_WDT*MAX_TEXT_HGT];



void refresh_engine_list(void)
{
	char filename[MAXPATH], engineDirectory[MAXPATH];

	strncpy(filename, baseDirectory, sizeof(filename));
	strncat(filename, "engines/list", sizeof(filename));

	strncpy(engineDirectory, baseDirectory, sizeof(engineDirectory));
	strncat(engineDirectory, "engines", sizeof(engineDirectory));

	FILE* file = xfopen(filename, "wt");

	if (!file) return;

	clear_screen(COL_PROGRAM_BACKGROUND);

	int y = 2;
	int engine_cnt = 0;
	DIR *dp;

	if ((dp = opendir(engineDirectory)))
	{
		struct dirent *dirp;

		while ((dirp = readdir(dp)))
		{
			if (dirp->d_type == DT_REG)
			{
				if (!strcmp(&dirp->d_name[strlen(dirp->d_name) - strlen(ENGINE_EXTENSION)], ENGINE_EXTENSION))
				{
					++engine_cnt;

					char str[1024];
					snprintf(str, sizeof(str), "Updating engine list (%i)...", engine_cnt);
					put_str(1, 1, str, COL_DIRECTORY);

					if (y < textScreenHeight - 2)
					{
						++y;
					}
					else
					{
						int scroll_y = 3;
						int scroll_hgt = textScreenHeight - scroll_y - 1;
						for (int i = 0; i < scroll_hgt; ++i)
						{
							if (i < scroll_hgt - 1)
							{
								memcpy(&textBuf[scroll_y*MAX_TEXT_WDT], &textBuf[(scroll_y + 1)*MAX_TEXT_WDT], textScreenWidth);
							}
							else
							{
								memset(&textBuf[scroll_y*MAX_TEXT_WDT], 32, textScreenWidth);
							}
							++scroll_y;
						}

					}

					put_str(3, y, dirp->d_name, COL_PROGRAM_BACKGROUND);

					update_screen();

					if (engine_get_info(dirp->d_name))
					{
						xfprintf(file, "%s\n", dirp->d_name);
						xfprintf(file, "%s\n", selectEngineTitle);
						xfprintf(file, "%s\n", selectEngineAbout);
						xfprintf(file, "###\n");
					}
				}
			}
		}

		closedir(dp);
	}

	xfclose(file);
}



void check_engine_updates(void)
{
	char hashname[MAXPATH], listname[MAXPATH], filename[MAXPATH], engineDirectory[MAXPATH], temp[1024];

	char* hashPrev = NULL;

	strncpy(engineDirectory, baseDirectory, sizeof(engineDirectory));
	strncat(engineDirectory, "engines", sizeof(engineDirectory));

	int hashPrevSize = 0;
	int hashNewSize = 65536;

	char* hashNew = (char*)malloc(hashNewSize);

	memset(hashNew, 0, hashNewSize);

	strncpy(hashNew, "", hashNewSize);

	DIR *dp;

	if ((dp = opendir(engineDirectory)))
	{
		struct dirent *dirp;

		while ((dirp = readdir(dp)))
		{
			if (dirp->d_type == DT_REG)
			{
				if (!strcmp(dirp->d_name, "hash") || !strcmp(dirp->d_name, "list")) continue;

				strncpy(filename, engineDirectory, sizeof(filename));
				strncat(filename, "/", sizeof(filename));
				strncat(filename, dirp->d_name, sizeof(filename));

				FILE* file = xfopen(filename, "rb");

				if (!file) continue;

				struct stat fileStat;

#ifdef WINDOWS
				fstat(_fileno(file), &fileStat);
#else
				fstat(fileno(file), &fileStat);
#endif
				xfclose(file);

				snprintf(temp, sizeof(temp), "%s %i %8.8x\n", dirp->d_name, fileStat.st_size, (unsigned int)fileStat.st_mtime);

				if ((int)(strlen(hashNew) + strlen(temp)) >= hashNewSize)
				{
					hashNewSize += 65536;
					hashNew = (char*)realloc(hashNew, hashNewSize);
				}

				strncat(hashNew, temp, hashNewSize);
			}
		}

		closedir(dp);
	}

	strncpy(hashname, engineDirectory, sizeof(hashname));
	strncat(hashname, "/hash", sizeof(hashname));
	strncpy(listname, engineDirectory, sizeof(hashname));
	strncat(listname, "/list", sizeof(hashname));

	bool refresh = true;

	FILE* file = xfopen(listname, "rb");

	if (file)
	{
		xfclose(file);

		file = xfopen(hashname, "rb");

		if (file)
		{
			xfseek(file, 0, SEEK_END);
			hashPrevSize = xftell(file);
			xfseek(file, 0, SEEK_SET);

			char* hashPrev = (char*)malloc(hashPrevSize);

			xfread(hashPrev, hashPrevSize, 1, file);
			xfclose(file);

			if (!strcmp(hashPrev, hashNew)) refresh = false;

			free(hashPrev);
		}
	}

	if (refresh)
	{
		file = xfopen(hashname, "wb");

		if (file)
		{
			xfwrite(hashNew, strlen(hashNew) + 1, 1, file);
			xfclose(file);
		}

		refresh_engine_list();
	}

	free(hashNew);
}



void free_select_engine_list(void)
{
	selectEngineFileList.clear();
	selectEngineTitleList.clear();
	selectEngineAboutList.clear();

	strncpy(selectEngineTitle, "", sizeof(selectEngineTitle));
	strncpy(selectEngineAbout, "", sizeof(selectEngineAbout));

	selectEngineCount = 0;
}



void select_engine_update_info(void)
{
	if (!selectEngineCount) return;

	strncpy(selectEngineTitle, selectEngineTitleList[selectEngineCursor].c_str(), sizeof(selectEngineTitle));
	strncpy(selectEngineAbout, selectEngineAboutList[selectEngineCursor].c_str(), sizeof(selectEngineAbout));
}



void set_select_engine(void)
{
	char listname[MAXPATH], filename[MAXPATH];
	FILE *file;
	int i;

	check_engine_updates();

	free_select_engine_list();

	strncpy(listname, baseDirectory, sizeof(listname));
	strncat(listname, "engines/list", sizeof(listname));

	file = xfopen(listname, "r");

	if (!file) critical_error("[SYS] Can't open engine list");

	while (1)
	{
		if (!xfgets(filename, sizeof(filename), file)) break;

		for (i = 0; i < (int)strlen(filename); ++i) if (filename[i] < 32) filename[i] = 0;

		xfgets(selectEngineTitle, sizeof(selectEngineTitle), file);

		for (i = 0; i < (int)strlen(selectEngineTitle); ++i) if (selectEngineTitle[i] < 32) selectEngineTitle[i] = 0;

		selectEngineFileList.push_back(filename);
		selectEngineTitleList.push_back(selectEngineTitle);

		strncpy(selectEngineAbout, "", sizeof(selectEngineAbout));

		while (1)
		{
			if (!xfgets(filename, sizeof(filename), file)) break;
			if (!memcmp(filename, "###", 3)) break;

			strncat(selectEngineAbout, filename, sizeof(selectEngineAbout));
		}

		selectEngineAboutList.push_back(selectEngineAbout);

		++selectEngineCount;
	}

	xfclose(file);

	selectEngineCursor = 0;

	for (i = 0; i < selectEngineCount; ++i)
	{
		if (!strcmp(engineFileName, selectEngineFileList[i].c_str()))
		{
			selectEngineCursor = i;
			break;
		}
	}

	select_engine_update_info();

	set_edit_mode(MODE_SELECT_ENGINE);
}



const char* get_engine_name(void)
{
	int off = 0;

	for (int i = (int)strlen(selectEngineTitle) - 1; i >= 0; --i)
	{
		if (selectEngineTitle[i] == '|')
		{
			off = i + 1;
			break;
		}
	}

	return &selectEngineTitle[off];
}



void draw_select_engine(void)
{
	char buf[MAX_TEXT_WDT];

	for (int i = 0; i < textScreenWidth; ++i)
	{
		PUT_ATTR(i, 0, COL_HEADER);
		PUT_ATTR(i, textScreenHeight - 1, COL_HEADER);
	}

	for (int i = 0; i < PATTERN_HEIGHT / 2 - 1; ++i)
	{
		for (int j = 0; j < textScreenWidth; ++j)
		{
			PUT_ATTR(j, PATTERN_YPOS + i, COL_PATTERN_BEAT);
		}
	}

	snprintf(buf, sizeof(buf), "Select engine (%i found)", selectEngineCount);

	put_str(1, 0, buf, COL_HEADER);
	put_str(1, textScreenHeight - 1, "~Ctrl+Enter~ change engine (unsaved data will be irreversibly changed!)", COL_HEADER);

	if (!selectEngineCount)
	{
		put_str(PATTERN_XPOS + 1, PATTERN_YPOS + 1, "No engines found", COL_CURSOR_ROW);

		return;
	}

	snprintf(buf, sizeof(buf), "Engine: %s (%s)", get_engine_name(), selectEngineFileList[selectEngineCursor].c_str());

	put_str(PATTERN_XPOS + 1, PATTERN_YPOS + 1, buf, COL_PATTERN_BEAT);

	put_str_wrap(PATTERN_XPOS + 1, PATTERN_YPOS + 3, textScreenWidth - 2, PATTERN_HEIGHT / 2 - 1, selectEngineAbout, COL_PATTERN_BEAT, 0);

	int y = PATTERN_HEIGHT / 2;
	int offset = 0;

	if (selectEngineCount < ENGINE_LIST_HEIGHT)
	{
		offset = 0;
	}
	else
	{
		offset = selectEngineCursor - ENGINE_LIST_HEIGHT / 2;

		if (offset > selectEngineCount - ENGINE_LIST_HEIGHT) offset = selectEngineCount - ENGINE_LIST_HEIGHT;

		if (offset < 0) offset = 0;
	}

	int offseti = offset;

	for (int i = 0; i < ENGINE_LIST_HEIGHT; ++i)
	{
		if (offset >= selectEngineCount) break;

		if (!strcmp(engineFileName, selectEngineFileList[offset].c_str())) put_str(PATTERN_XPOS + 2, y, "->", COL_DIRECTORY);

		int color = (offset == selectEngineCursor) ? COL_CURSOR_ROW : COL_PROGRAM_BACKGROUND;

		const char* str = selectEngineTitleList[offset].c_str();

		int x = PATTERN_XPOS + 4;

		put_str(x, y, offset == selectEngineCursor ? ">" : " ", color);

		++x;

		int xs = 0;

		for (int j = 0; j < (int)strlen(str); ++j)
		{
			if (str[j] == '|')
			{
				if (x < PATTERN_XPOS + 5 + 5) xs = PATTERN_XPOS + 5 + 5; else xs = PATTERN_XPOS + 5 + 11;

				while (x < xs)
				{
					PUT_CHAR(x, y, ' ');
					PUT_ATTR(x, y, color);

					++x;
				}

				continue;
			}

			PUT_CHAR(x, y, str[j]);
			PUT_ATTR(x, y, color);

			++x;
		}

		put_str(x, y, " ", color);

		++y;
		++offset;
	}

	char cu = ' ';
	char cd = ' ';

	if (offseti > 0) cu = 16;
	if (offseti < (selectEngineCount - ENGINE_LIST_HEIGHT)) cd = 17;

	PUT_CHAR_ATTR(PATTERN_XPOS + 1, PATTERN_HEIGHT / 2 + 1, cu, COL_DIRECTORY);
	PUT_CHAR_ATTR(PATTERN_XPOS + 1, PATTERN_HEIGHT / 2 + 2, cu, COL_DIRECTORY);
	PUT_CHAR_ATTR(PATTERN_XPOS + 1, textScreenHeight - 4, cd, COL_DIRECTORY);
	PUT_CHAR_ATTR(PATTERN_XPOS + 1, textScreenHeight - 3, cd, COL_DIRECTORY);
}



void move_select_engine_cursor(int offset)
{
	selectEngineCursor += offset;

	if (selectEngineCursor < 0) selectEngineCursor = 0;
	if (selectEngineCursor >= selectEngineCount - 1) selectEngineCursor = selectEngineCount - 1;

	select_engine_update_info();
}



void select_new_engine(void)
{
	strncpy(engineFileName, selectEngineFileList[selectEngineCursor].c_str(), sizeof(engineFileName));

	set_undo();
	module_clear();
	change_engine();

	int noteSrcColumn = 0;

	for (int i = 0; i < moduleUndo.columns; ++i)
	{
		if (moduleUndo.columnType[i] == COLUMN_TYPE_NOTE)
		{
			int noteDstColumn = 0;

			for (int j = 0; j < module.columns; ++j)
			{
				if (module.columnType[j] == COLUMN_TYPE_NOTE)
				{
					if (noteDstColumn == noteSrcColumn)
					{
						for (int k = 0; k < MAX_ROWS; ++k)
						{
							int note = moduleUndo.rows[activeTrack][k].column[i];

							if (note > 1)
							{
								if (note < module.lowestNote) note = module.lowestNote;
								if (note > module.highestNote) note = module.highestNote;
							}

							module.rows[activeTrack][k].column[j] = note;
						}

						break;
					}

					++noteDstColumn;
				}
			}

			++noteSrcColumn;
		}
	}

	for (int i = 0; i < MAX_ROWS; ++i)
	{
		module.marker[i] = moduleUndo.marker[i];
		strncpy(module.markerName[i], moduleUndo.markerName[i], sizeof(module.markerName[i]));
	}

	module.loopStart = moduleUndo.loopStart;
	module.loopEnd = moduleUndo.loopEnd;
	//module.speed    =moduleUndo.speed;	//use default speed of the newly selected engine
	module.measure = moduleUndo.measure;

	strncpy(module.title, moduleUndo.title, sizeof(module.title));
	strncpy(module.author, moduleUndo.author, sizeof(module.author));

	set_edit_mode(MODE_EDIT_SONG);
}



void select_engine_events(SDL_Event event)
{
	switch (event.type)
	{
	case SDL_KEYDOWN:
	{
		switch (event.key.keysym.sym)
		{
		case SDLK_ESCAPE:
		{
			set_edit_mode(MODE_EDIT_SONG);
		}
		break;

		case SDLK_UP:
		{
			move_select_engine_cursor(-1);
		}
		break;

		case SDLK_DOWN:
		{
			move_select_engine_cursor(1);
		}
		break;

		case SDLK_PAGEUP:
		{
			move_select_engine_cursor(-10);
		}
		break;

		case SDLK_PAGEDOWN:
		{
			move_select_engine_cursor(10);
		}
		break;

		case SDLK_HOME:
		{
			selectEngineCursor = 0;
			select_engine_update_info();
		}
		break;

		case SDLK_END:
		{
			selectEngineCursor = selectEngineCount - 1;
			select_engine_update_info();
		}
		break;

		case SDLK_RETURN:
		{
			if (event.key.keysym.mod&(KMOD_LCTRL | KMOD_RCTRL))
			{
				select_new_engine();
			}
		}
		break;
		}
	}
	break;
	}
}